/* Copyright (C) 1998-1999 Kimmo.palin@helsinki.fi
   License: see COPYING
*/
#include<gtk/gtk.h>
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<assert.h>
#include<sys/stat.h>
#include<unistd.h>
#include"ttk91vm.h"
/*#include"y.tab.h"*/
#include"mem_access.h"
#include"compiler.h"
#include"cpu.h"



#define EI_VOI_LUKEA "En voi lukea tiedostoa %s"

extern FILE *yyin;

syment SymTable={NULL,SYMUNKNOWN,0,0,NULL,NULL};
cpudev *boot_cpu;


void reset_compiler()
{
  syment *this=SymTable.next,*prev;
  rlent *rlthis,*rlprev;

  while(this!=NULL) {
    rlthis=this->rltable;
    while(rlthis!=NULL) {
      rlprev=rlthis;
      rlthis=rlthis->next;
      free(rlprev);
    }
    prev=this;
    this=this->next;
    free(prev);
  }
  SymTable.next=NULL;
  SymTable.rltable=NULL;
  progcount=memlimit=0;
  line_count=1;

  /* Flush flex-buffer*/
  reset_lexer(yyin);

}

void gtk_add_symbols()
{
  syment *this=SymTable.next;
  gchar *tmp[2],value[32];
  
  tmp[1]=value;
  
  gtk_clist_freeze(GtkSymList);
  gtk_clist_clear(GtkSymList);

  while(this!=NULL) {
    if(this->type!=SYMSYSCONST) {
      tmp[0]=this->name;
      sprintf(value,"%d",this->value);
      gtk_clist_append(GtkSymList,tmp);
    }
    this=this->next;
  }
  gtk_clist_thaw(GtkSymList);
}

void print_symboltable()
{
  syment *this=SymTable.next,*prev;

  while(this!=NULL) {
    fprintf(stderr,"%s,%d,%d\n",this->name,this->type,this->value);
    prev=this;
    this=this->next;
    free(prev);
  }
}
  

syment *find_symbol(char *name)
{
  syment *this=SymTable.next;

#ifdef MATCHCASE
  while(this!=NULL && strcmp(this->name,name)!=0)
#else
  while(this!=NULL && strcasecmp(this->name,name)!=0)
#endif
      this=this->next;

  return this;
}

/* Lis symbolin symbolitauluun ilman mitn tarkistuksia */

syment *create_symbol(char *name, int type, bit32 value)
{
  syment *inst;
  
  inst=(syment*)malloc(sizeof(syment));

  inst->rltable=NULL;
  inst->name=(char*)malloc(strlen(name));
  strcpy(inst->name,name);

  inst->type=type;
  inst->value=value;

  inst->next=SymTable.next;
  SymTable.next=inst;

  return inst;
}

/* Lis ennen kuulumattoman symbolin tai asettaa uudet arvot.
 palauttaa:  0 jos onnistuu
 -1 jos yritetn asettaa uudestaan SYMSYSCONST tyyppinen symboli*/
int add_symbol(char *name,int type,bit32 value,int line)
{
  syment *inst;

  inst=find_symbol(name);

  if(inst!=NULL) {
    if(inst->type==SYMSYSCONST)
      return -1;
    inst->type=type;
    inst->value=value;
  } else {
    inst=create_symbol(name,type,value);
  }
  inst->def_line=line;
  return 0;
}

/* Lis paikan uudelleensijoitustauluun */
void addrl_symbol(syment *ent, bit32 progloc)
{
  rlent *add=(rlent*)malloc(sizeof(rlent));

  add->rlpos=progloc;
  add->next=ent->rltable;
  ent->rltable=add;
}


/* Palauttaa symbolin arvon. Jos symbolia ei ole ennen luotu niin luo sen,
   lis tmn kohdan kyseisen symbolin uudelleensijoitustauluun
   HUOM! Kaikki kohdat merkitn rltableen!!! */
bit32 get_symbol(char *name,bit32 progloc)
{
  syment *tmp=find_symbol(name);

  if(tmp==NULL)
    tmp=create_symbol(name,SYMUNKNOWN,0);

  addrl_symbol(tmp,progloc);
  
  return tmp->value;
}


/* Hoitelee uudelleensijoituslistan yhden symbolin kohdalta.
   Palauttaa uudellensijoitusten mrn*/
int procRlList(syment *sym)
{
  rlent *base=sym->rltable;
  int relocs=0;
  memunit rl;

  while(base!=NULL) {
    rl=getMem(boot_cpu,base->rlpos);
    rl.flds.addr=sym->value;
    setMem_with_sym(boot_cpu,base->rlpos,rl,sym->name);

    

    base=base->next;
    relocs++;
  }
  return relocs;
}

/*Hoitaa uudelleensijoituksen.
  Palauttaa: 0 jos onnistui
  -1 jos oli mrittelemttmi symboleita*/

int reLocate()
{
  syment *process=SymTable.next;
  int tmpaddr;

  while(process!=NULL) {
    switch(process->type) {
    case SYMDCONST:
      setMem(boot_cpu,memlimit,(memunit)process->value);
      process->value=memlimit;
      memlimit++;
      break;
    case SYMDSEQ:
      tmpaddr=memlimit;
      memlimit+=process->value;
      process->value=tmpaddr;
      break;
    case SYMEQU:
    case SYMADDR:
    case SYMSYSCONST:
    default:
      break;
    }
    if(process->def_line==0) {
      fprintf(stderr,"VIRHE!!! Kytetty mrittelemtnt muuttujaa %s!\n",
	      process->name);
      return -1;
    }

    if(procRlList(process)==0) {
      if(process->type!=SYMSYSCONST)
	set_message("Varoitus: Mritelty symboleita joita ei kytet!");
    }
    
    process=process->next;
  }

  return 0;
}

int comp_file(cpudev *cpu,const char *nimi)
{
  char *viesti;
  int parseret;
  struct stat f_status;

  boot_cpu=cpu;

  reset_compiler();

  if(nimi==NULL) {
    return -1;
  } else if((yyin=fopen(nimi,"r"))==NULL || 
	    /* Tarkistetaan ett kyseess on tavallinen tiedosto */
	    stat(nimi,&f_status)==-1 || !S_ISREG(f_status.st_mode)) {
    viesti=(char*)malloc(strlen(EI_VOI_LUKEA)+strlen(nimi)+2);
    sprintf(viesti,EI_VOI_LUKEA,nimi);
    err_message(viesti);
    free(viesti);
    return -2;
  }
  reset_lexer(yyin);

  add_symbol("CRT",SYMSYSCONST,D_CRT,1);
  add_symbol("KBD",SYMSYSCONST,D_KBD,1);
  add_symbol("HALT",SYMSYSCONST,S_HALT,1);

  gtk_clist_freeze(GtkMemList);
  parseret=yyparse();
  gtk_clist_thaw(GtkMemList);

  memlimit=progcount;

  if(yyin!=stdin) {
    fclose(yyin);
    yyin=NULL;
  }

  if(parseret!=1) {
    parseret=reLocate();
    if(parseret!=0)
      set_message("Kytetty mrittelemttmi symboleita!");
    else
      set_message("Knns onnistui");

    return parseret;
  } else {
    set_message("Syntaksivirheit!!");
    return parseret;
  }
    
}
