/* Copyright (C) 1998-1999 Kimmo.palin@helsinki.fi
   License: see COPYING
*/

#include<assert.h>
#include<stdlib.h>
#include<limits.h>
#include"mem_access.h"
#include"cpu.h"
#include"cpu_access.h"

static memunit *core;
static int coreSize;

gchar *regnames[]={"R0","R1","R2","R3","R4","R5","SP","FP"};
char *memloads="= @";


void mem_select(int addr)
{
  gtk_clist_select_row(GtkMemList,addr,0);
}

void sinstr(gchar *buf,opdata comm)
{
  sprintf(buf,"%s %s,%c%d",asmOp(NULL,comm.opcode),regnames[comm.rj],
	  memloads[comm.mem],comm.addr);
  if(comm.ri!=0) {
    sprintf(buf+strlen(buf),"(%s)",regnames[comm.ri]);
  }
}

void sinstr_with_sym(gchar *buf,opdata comm,gchar *sym_name)
{
  sprintf(buf,"%s %s,%c%s",asmOp(NULL,comm.opcode),regnames[comm.rj],
	  memloads[comm.mem],sym_name);
  if(comm.ri!=0) {
    sprintf(buf+strlen(buf),"(%s)",regnames[comm.ri]);
  }
}


int init_core(int size)
{
  int i;
  gchar *tmp[3],addr[16],val[16],instr[32];

  tmp[0]=addr;
  tmp[1]=val;
  tmp[2]=instr;

  core=(memunit*)calloc(sizeof(memunit),size);

  assert(core!=NULL);
  coreSize=size;
  gtk_clist_freeze(GtkMemList);
  for(i=0;i<coreSize;i++) {
    sprintf(addr,"%d",i);
    sprintf(val,"%d",core[i].bin);
    sinstr(instr,core[i].flds);
    gtk_clist_append(GtkMemList,tmp);
  }
  gtk_clist_thaw(GtkMemList);
  return coreSize;
}

int getCoreSize()
{
  return coreSize;
}


/* Asettaa arvon muistiin mutta ei hoitele kyttliittym. */
int setMem_no_ui(cpudev *cpu, bit32 addr, memunit value)
{
  if((get_sr(cpu)&STAT_P) && (addr<0 || addr>=get_limit(cpu)) ) {
    or_sr(cpu,STAT_M);
    return -1;
  }

  set_mar(cpu,get_base(cpu)+addr);
  set_mbr(cpu,value);

  core[get_mar(cpu)]=get_mbr(cpu);
  return 0;
}

int setMem_with_sym(cpudev *cpu, bit32 addr, memunit value,gchar *sym_name)
{
  gchar val[16],instr[300];

  setMem_no_ui(cpu,addr,value);


  sprintf((char*)val,"%d",value.bin);
  sinstr_with_sym(instr,core[get_mar(cpu)].flds,sym_name);

  gtk_clist_set_text(GtkMemList,addr,1,val);
  gtk_clist_set_text(GtkMemList,addr,2,instr);

  mem_select(addr);
  return 0;
}

int setMem(cpudev *cpu, bit32 addr, memunit value)
{
  gchar val[16],instr[32];

  setMem_no_ui(cpu,addr,value);


  sprintf((char*)val,"%d",value.bin);
  sinstr(instr,core[get_mar(cpu)].flds);
  gtk_clist_set_text(GtkMemList,addr,1,val);
  gtk_clist_set_text(GtkMemList,addr,2,instr);

  mem_select(addr);
  return 0;
}
/* Virheilmoitus!!! hoida errno muuttujalla*/
memunit getMem(cpudev *cpu,bit32 addr)
{
  memunit retunit;

  if((get_sr(cpu)&STAT_P) && (addr<0 || addr>=get_limit(cpu)) ) {
    or_sr(cpu,STAT_M);
    retunit.bin=INT_MIN;
    return retunit;         /*ANSI-C:n vastainen. Muutenkin korjattava*/
  }

  /*  set_mar(cpu,get_base(cpu)+addr);
  retunit=get_mbr(cpu);
  */
  
  mem_select(addr);
  return core[addr];
}



/* For debuging */
void print_core()
{
  opdata *tmp=(opdata*)core;
  int i;
  char opname[20];

  for(i=0;i<progcount;i++) 
    fprintf(stderr,"%d: (%d) \t%s R%d,%d %d(R%d)\n",i,tmp[i].opcode,
	    asmOp(opname,tmp[i].opcode),tmp[i].rj,tmp[i].mem,
	    tmp[i].addr,tmp[i].ri);
}




